#include <abort.h>
#include <malloc.h>
#include <string.h>
#include <containers/vec_.h>

Vec* new_Vec( size_t el_size, size_t alignment ) {
	Vec* vec = malloc( sizeof(Vec) );
	if(NULL == vec)
		abort();
	init_Vec( vec, el_size, alignment );
	return vec;
}

inline void delete_Vec( Vec* self ) {
	destroy_Vec( self );
	free( self );
}

void init_Vec( Vec* self, size_t el_size, size_t alignment ) {
	self->el_size = el_size;
	self->alignment = alignment;
	self->num_el = 0;
	self->max_el = 16;
	if( NULL == (
		self->data_unaligned =
		 	malloc( el_size*self->max_el + (1<<alignment) )
		)
	) {
		abort();
	}
	self->end = self->data = ALIGN_ADR( self->data_unaligned, alignment );
}

inline void destroy_Vec( Vec* self ) {
	free( self->data_unaligned );
}

void resize_Vec( Vec* self, uint32_t newsize ) {
	if( NULL == (
		self->data_unaligned =
			realloc( self->data_unaligned, self->el_size*newsize + (1<<self->alignment) )
		)
	) {
		self->data = NULL;
		self->end = NULL;
		abort();
	}
	self->data = ALIGN_ADR( self->data_unaligned, self->alignment );
	self->end = (uint8_t*)self->data + self->el_size * self->num_el;
	self->max_el = newsize;
}

void pushBack_Vec( Vec* self, void const* element ) {
	if( self->num_el >= self->max_el ) {
		resize_Vec( self, self->max_el << 1 );
	}

	memcpy( self->end, element, self->el_size );
	self->end = (uint8_t*)self->end + self->el_size;
	self->num_el++;
}

inline void* getElement_Vec( Vec* self, uint32_t i ) {
	return (uint8_t*)self->data + i * self->el_size;
}

inline void const* getConstElement_Vec( Vec const* self, uint32_t i ) {
	return (uint8_t*)self->data + i * self->el_size;
}

inline void* getHead_Vec( Vec* self ) {
	return self->data;
}

inline void* getTail_Vec( Vec* self ) {
	return (uint8_t*)self->data + self->num_el * self->el_size;
}

inline uint32_t getLength_Vec( Vec const* self ) {
	return self->num_el;
}

inline size_t getElementSize_Vec( Vec const* self ) {
	return self->el_size;
}

inline void reset_Vec( Vec* self ) {
	self->num_el = 0;
	self->end = self->data;
}

inline VecIterator createIterator_Vec( Vec* self ) {
	VecIterator it;
	init_VecIterator( &it, self );

	return it;
}

inline void init_VecIterator( VecIterator* self, Vec* vec ) {
	self->vec = vec;
	self->current_idx = 0;
	self->num_elements = getLength_Vec( vec );
	self->current_element = getHead_Vec( vec );
}

inline void destroy_VecIterator( VecIterator* self ) {
	if( self ) {}
}

inline void* dereference_VecIterator( VecIterator* self ) {
	if( self->current_idx < self->num_elements ) {
		void* element = self->current_element;
		return element;
	} else {
		return NULL;
	}
}

inline bool next_VecIterator( VecIterator* self ) {
	if( self->current_idx < self->num_elements ) {
		self->current_element = (void*)((uintptr_t)self->current_element + getElementSize_Vec( self->vec ));
		self->current_idx++;
		return true;
	} else {
		return false;
	}
}
