#include <abort.h>
#include <float.h>
#include <engine.h>
#include <rectangle.h>
#include <sys/region.h>
#include <sys/sys.h>
#include <node/xformer_funcs.h>

#define MAX_NODES	64

static bool static_mesh_is_opaque( StaticMesh const* staticMesh ) {
	return 1.0f <= staticMesh->opacity;
}

void init_Engine( Engine* self ) {
	init_DList( &self->dl );

	init_Vec( &self->lights, sizeof(Lux*), ALIGN_DEFAULT );
	init_Vec( &self->materials, sizeof(Material*), ALIGN_DEFAULT );

	init_Vec( &self->cpuMatrixes, sizeof(Matrix), ALIGN_4B );
	resize_Vec( &self->cpuMatrixes, MAX_NODES );

	init_Vec( &self->xformers, sizeof(Xformer), ALIGN_DEFAULT ); 	// how to get a structure's alignment?
	resize_Vec( &self->xformers, 30 );								// don't mess with the vec from now - we don't want any reallocations to occur

	init_Vec( &self->static_meshes, sizeof(StaticMesh), ALIGN_DEFAULT );
	resize_Vec( &self->static_meshes, 30 );

	init_Vec( &self->meshes, sizeof(Mesh), ALIGN_DEFAULT );
	resize_Vec( &self->meshes, 30 );

	init_Vec( &self->billboards, sizeof(Billboard), ALIGN_DEFAULT );
	resize_Vec( &self->billboards, 30 );

	init_Vec( &self->cameras, sizeof(Camera), ALIGN_DEFAULT );
	resize_Vec( &self->cameras, 30 );

	init_Vec( &self->luxes, sizeof(Lux), ALIGN_DEFAULT );
	resize_Vec( &self->luxes, 30 );

	init_Vec( &self->flares, sizeof(Flare), ALIGN_DEFAULT );
	resize_Vec( &self->flares, 30 );

	init_Vec( &self->transparentXf, sizeof(Raw_Xformer*), ALIGN_DEFAULT );
	init_Vec( &self->opaqueXf, sizeof(Raw_Xformer*), ALIGN_DEFAULT );
	init_Vec( &self->flaresXf, sizeof(Flare_Xformer*), ALIGN_DEFAULT );

	self->camera_2d_xf = createCamera_Engine( self, ORTHOGRAPHIC_CAMERA );

	Viewport vp;
	init_Viewport( &vp, BUFFER_W, BUFFER_H, G_MAXZ/2, G_MAXZ/2 );			// TODO: make getters in Sys module for these values
	setViewport_Engine( self, &vp );
}

inline void setViewport_Engine( Engine* self, Viewport const* viewport ) {
	self->viewport = clone_Viewport( viewport );
}

inline void setCamera_Engine( Engine* self, Camera_Xformer const* camera_xf ) {
	self->camera_xf = camera_xf;
}

inline void setRenderTarget_Engine( Engine* self, Surface const* renderTarget ) {
	self->renderTarget = renderTarget;
}

inline void setZBuffer_Engine( Engine* self, Surface const* zbuffer ) {
	self->zBuffer = zbuffer;
}

inline Viewport const* getViewport_Engine( Engine const* self ) {
	return &self->viewport;
}

inline Camera_Xformer const* getCamera_Engine( Engine const* self ) {
	return self->camera_xf;
}

inline Surface const* getRenderTarget_Engine( Engine const* self ) {
	return self->renderTarget;
}

inline Surface const* getZBuffer_Engine( Engine const* self ) {
	return self->zBuffer;
}

inline DList* getDisplayList_Engine( Engine* self ) {
	return &self->dl;
}

static inline Matrix* getProjectionMatrix_Engine( Engine* self ) {
	return &self->projMtxF;
}

inline void addLight_Engine( Engine* self, Lux const* light ) {
	pushBack_Vec( &self->lights, &light );
}

inline void addMaterial_Engine( Engine* self, Material const* material ) {
	pushBack_Vec( &self->materials, &material );
}

inline void removeLight_Engine( Engine* self, Lux const* light ) {
	if( self ) {}
	if( light ) {}
}

static void handleStaticMesh( DList* dlist, StaticMesh_Xformer const* staticMesh_xf, Camera_Xformer const* camera_xf ) {
	pushLoadMatrix_DList( dlist,
		getResultMatrix_Xformer( staticMesh_xf->xf ),
		MATRIX_MODELVIEW
	);

	CombineMode prev_cm = -1;
	CombineModeAlpha prev_cma = -1;
	RenderMode prev_rm = -1;

	StaticMesh const* staticMesh = staticMesh_xf->static_mesh;
	StaticLayer const* layer = staticMesh->layers;
	for( uint32_t i=0, n=staticMesh->num_layers; i<n; ++i, ++layer ) {
		Material mat;
		clone_Material( &staticMesh->materials[ layer->material_id ],
			&mat
		);
		getColor_Material( &mat ) -> t *= staticMesh->opacity;
		pushSetMaterial_DList( dlist, &mat );

		// calculate combine and rendering modes
		CombineMode cm;
		CombineModeAlpha cma;
		RenderMode rm;
		if( TEXTURE_ID_NONE != getTextureId_Material( &mat ) ) {
			Texture* tex = &staticMesh->textures[ getTextureId_Material( &mat ) ];
			Rectangle tttile = {
				0,
				0,
				getWidth_Texture( tex ) - 1,
				getHeight_Texture( tex ) - 1
			};

			Rectangle const* tile = &layer->tile;
			if( (tile->p0x == 0) && (tile->p0y == 0) && (tile->p1x == 0) && (tile->p0y == 0) ) {		// HACKY
				pushLoadTexture_DList( dlist, tex, tttile, false );
			} else {
				pushLoadTexture_DList( dlist, tex, *tile, true );
			}

			switch( getTextureCombine_Material( &mat ) ) {
				case TEXTURECOMBINE_DECAL: {
					cm = CM_DECAL;
					cma = (getColor_Material( &mat )->t == 1.0f) ? CMA_OPAQUE : CMA_TRANSLUCENT;
					rm = (getColor_Material( &mat )->t == 1.0f) ? RM_ZB_TEX_EDGE : RM_ZB_TRANSLUCENT;	// ? ZB_OPAQUE
					break;
				}
				case TEXTURECOMBINE_ALPHA: {
					cm = CM_DECAL;				//G_CC_BLENDI;
					cma = CMA_TRANSLUCENT;
					rm = RM_ZB_TRANSLUCENT;
					break;
				}
				default: {
					abort();
				}
			}

			if( TEXTUREMAPPING_ENV == getTextureMapping_Material( &mat ) ) {
				pushSetTextureScale_DList( dlist, 0.125f, 0.125f );
				pushSetGeometryMode_DList( dlist, GM_LIGHTING | GM_TEXGEN );
				Coordsys const* camcs = getCoordsys_Xformer( camera_xf->xf );
				Coordsys const* meshcs = getCoordsys_Xformer( staticMesh_xf->xf );
				
				pushSetLookAt_DList( dlist, &camcs->origin, &meshcs->origin, &camcs->v );
			}
		} else {
			cm = CM_SHADE;
			cma = (getColor_Material( &mat )->t == 1.0f) ? CMA_OPAQUE : CMA_TRANSLUCENT;
			rm = (getColor_Material( &mat )->t == 1.0f) ? RM_ZB_TEX_EDGE : RM_ZB_TRANSLUCENT;	// ? ZB_OPAQUE
		}

		if( prev_rm != rm ) {
			pushSetRenderMode_DList( dlist, rm );
			prev_rm = rm;
		}
		if( prev_cm != cm || prev_cma != cma ) {
			pushSetCombineMode_DList( dlist, cm, cma );
			prev_cm = cm;
			prev_cma = cma;
		}

		pushSetSegment_DList( dlist, 0x0, (void*)layer->vts_ofs );
		pushJumpToTranslated_DList( dlist, layer->gfx );
	}
}

static void handleMesh( DList* dlist, Mesh_Xformer const* mesh_xf, Camera_Xformer const* camera_xf, Material const* materials[] ) {
	if( camera_xf ) {}

	pushLoadMatrix_DList( dlist,
		getResultMatrix_Xformer( mesh_xf->xf ),
		MATRIX_MODELVIEW
	);
	pushSetVertexScale_DList( dlist, 200.0f );

	CombineMode prev_cm = -1;
	CombineModeAlpha prev_cma = -1;
	RenderMode prev_rm = -1;

	Mesh const* mesh = mesh_xf->mesh;
	Vec const* layers = getLayersConst_Mesh( mesh );
	for( uint32_t i=0; i<getLength_Vec( layers ); ++i ) {
		Layer const* layer = getConstElement_Vec( layers, i );

		Material material;
		clone_Material( materials[ getMaterialId_Layer( layer ) ],
			&material
		);
		getColor_Material( &material )->t *= getOpacity_Mesh( mesh );
		pushSetMaterial_DList( dlist, &material );

		CombineMode cm = CM_SHADE;
		CombineModeAlpha cma = (getColor_Material( &material )->t == 1.0f) ? CMA_OPAQUE : CMA_TRANSLUCENT;
		RenderMode rm = (getColor_Material( &material )->t == 1.0f) ? RM_ZB_TEX_EDGE : RM_ZB_TRANSLUCENT;	// ? ZB_OPAQUE
		if( prev_rm != rm ) {
			pushSetRenderMode_DList( dlist, rm );
			prev_rm = rm;
		}
		if( prev_cm != cm || prev_cma != cma ) {
			pushSetCombineMode_DList( dlist, cm, cma );
			prev_cm = cm;
			prev_cma = cma;
		}

		Vec const* verts = getVerticesConst_Layer( layer );
		pushSetVertexSource_DList( dlist,
			getConstElement_Vec( verts, 0 ),
			0
		);
		pushLoadVertexBuffer_DList( dlist,
			0,
			0,
			getLength_Vec( verts )
		);

		Vec const* triangles = getTrianglesConst_Layer( layer );
		pushSetTriangleSource_DList( dlist,
			getConstElement_Vec( triangles, 0 )
		);

		pushDrawPrimitives_DList( dlist,
			PrimTriangle,
			getLength_Vec( triangles ),
			i*3
		);
	}
}

static void handleBillboard( DList* dlist, Billboard_Xformer const* bb_xf, Camera_Xformer const* camera_xf ) {
	handleStaticMesh( dlist, (StaticMesh_Xformer*)bb_xf, camera_xf );
}

static void resetState( DList* dlist ) {
	pushClearGeometryMode_DList( dlist, GM_LIGHTING | GM_TEXGEN );
	pushSetGeometryMode_DList( dlist,
		GM_CULLBACK |
		GM_ZBUFFER |
		GM_LIGHTING |
		GM_SHADE |
		GM_GOURAUD
	);
	pushSetTextureScale_DList( dlist, 1.0f, 1.0f );
	pushPipeSync_DList( dlist );
	pushSetRenderMode_DList( dlist, RM_ZB_OPAQUE );
}

static void handleProjection( DList* dlist, Matrix* projMtx, Viewport const* vp, Camera_Xformer const* cam_xf ) {
	switch( getType_Camera( cam_xf->camera ) ) {
		case PERSPECTIVE_CAMERA: {
			PerspCamera const* perspCam = getPersp_Camera( cam_xf->camera );
			float perspNorm;
			perspective_Matrix( projMtx,
				&perspNorm,
				getFOVY_PerspCamera( perspCam ),
				getWidth_Viewport( vp ) / getHeight_Viewport( vp ),
				getNear_PerspCamera( perspCam ),
				getFar_PerspCamera( perspCam )
			);
			pushNormalizePerspective_DList( dlist, perspNorm );

			break;
		}
		case ORTHOGRAPHIC_CAMERA: {
			OrthoCamera const* orthoCam = getOrtho_Camera( cam_xf->camera );
			orthographic_Matrix( projMtx,
				getLeft_OrthoCamera( orthoCam ),
				getRight_OrthoCamera( orthoCam ),
				getBottom_OrthoCamera( orthoCam ),
				getTop_OrthoCamera( orthoCam ),
				getNear_OrthoCamera( orthoCam ),
				getFar_OrthoCamera( orthoCam )
			);
			break;
		}
		default:
			abort();
			break;
	}

	Matrix inv;
	invert_Matrix( getMatrix_Xformer( cam_xf->xf ), &inv );
	multiply_Matrix( projMtx, &inv, projMtx );
	pushLoadMatrix_DList( dlist, projMtx, MATRIX_PROJECTION );
}

static void collectFlares( Vec* flaresXf, TreeNode const* tn_list[] ) {
	for( TreeNode const** it = tn_list; NULL != *it; ++it ) {
		Raw_Xformer const* raw_xf = getData_TreeNode( *it );
		switch( raw_xf->tag ) {
			case OBJECT_FLARE: {
				pushBack_Vec( flaresXf, &raw_xf );
				break;
			}
			default:
				break;
		}
	}
}

static void splitOpaqueTransparent( Vec* opaqueXf, Vec* transparentXf, TreeNode const* tn_list[] ) {
	for( TreeNode const** it = tn_list; NULL != *it; ++it ) {
		Raw_Xformer const* raw_xf = getData_TreeNode( *it );
		switch( raw_xf->tag ) {
			case OBJECT_BILLBOARD: {
				if( static_mesh_is_opaque( &( (Billboard_Xformer*)raw_xf )->billboard->mesh ) ) {
					pushBack_Vec( opaqueXf, &raw_xf );
				} else {
					pushBack_Vec( transparentXf, &raw_xf );
				}
				break;
			}
			case OBJECT_STATICMESH: {
				if( static_mesh_is_opaque( ( (StaticMesh_Xformer*)raw_xf )->static_mesh ) ) {
					pushBack_Vec( opaqueXf, &raw_xf );
				} else {
					pushBack_Vec( transparentXf, &raw_xf );
				}
				break;
			}
			case OBJECT_MESH: {
				pushBack_Vec( opaqueXf, &raw_xf );
				break;
			}
			default:
				break;
		}
	}
}

static void calculateWorldCoordinates( Matrix cpuMtx[], TreeNode const* tn_list[] ) {
	for( TreeNode const** it = tn_list; NULL != *it; ++it ) {
		Xformer* parent_xf = NULL;
		TreeNode const* parent = getParent_TreeNode( *it );
		if( parent ) {
			parent_xf = ( (Raw_Xformer*)getData_TreeNode( parent ) )->xf;
		}
		finish_Xformer(
			( (Raw_Xformer*)getData_TreeNode( *it ) )->xf,
			parent_xf,
			cpuMtx
		);
		++cpuMtx;
	}
}

static void handleLighting( DList* dlist, TreeNode const* tn_list[] ) {
	for( TreeNode const** it = tn_list; NULL != *it; ++it ) {
		Raw_Xformer const* raw_xf = getData_TreeNode( *it );

		switch( raw_xf->tag ) {
			case OBJECT_LUX: {
				pushLoadLight_DList( dlist,
					( (Lux_Xformer*)raw_xf )->lux,
					&getCoordsys_Xformer( raw_xf->xf )->origin,
					//&getCoordsys_Xformer( raw_xf->xf )->n,
					0
				);
				break;
			}
			default:
				break;
		}
	}
}

static void applyAdditionalTransformations( Camera_Xformer const* cam_xf, TreeNode const* tn_list[] ) {
	for( TreeNode const** it = tn_list; NULL != *it; ++it ) {
		Raw_Xformer const* raw_xf = getData_TreeNode( *it );
		void* data = NULL;
		switch( raw_xf->tag ) {
			case OBJECT_BILLBOARD: {
				data = cam_xf->xf;
				break;
			}
			default:
				break;
		}

		process_Xformer( raw_xf->xf, data );
		reset_Xformer( raw_xf->xf );
	}
}

static void buildSceneDL( DList* dlist, Camera_Xformer const* cam_xf, Vec const* vec[], uint32_t const vn, Vec const* materials  ) {
	for( uint_t vi=0; vi<vn; ++vi ) {
		for( uint32_t i=0; i<getLength_Vec( vec[vi] ); ++i ) {
			resetState( dlist );

			Raw_Xformer const* raw_xf = *(Raw_Xformer const**)getConstElement_Vec( vec[vi], i );
			switch( raw_xf->tag ) {
				case OBJECT_BILLBOARD:
					handleBillboard( dlist, (Billboard_Xformer const*)raw_xf, cam_xf );
					break;
				case OBJECT_STATICMESH: {
					handleStaticMesh( dlist, (StaticMesh_Xformer const*)raw_xf, cam_xf );
					break;
				}
				case OBJECT_MESH: {
					handleMesh(
						dlist,
						(Mesh_Xformer const*)raw_xf,
						cam_xf,
						(Material const**)getConstElement_Vec( materials, 0 )
					);
					break;
				}
				default:
					break;
			}
		}
	}
}

static void flattenTree( TreeNode* tree, TreeNode const* tn_list[] ) {
	TreeNodeIterator ni;
	init_TreeNodeIterator( &ni );
	reset_TreeNodeIterator( &ni, tree );

	int i = 0;
	for( TreeNode* tn; NULL != (tn = inc_TreeNodeIterator( &ni )); ++i ) {
		tn_list[i] = tn;
	}
	tn_list[i] = NULL;

	destroy_TreeNodeIterator( &ni );
}

void sceneTaskF_Engine( Engine* self, TreeNode* scene ) {

	// gather TreeNode pointers into a list
	static TreeNode const* tn_list[1+MAX_NODES];
	flattenTree( scene, tn_list );

	handleProjection(
		getDisplayList_Engine( self ),
		getProjectionMatrix_Engine( self ),
		getViewport_Engine( self ),
		getCamera_Engine( self )
	);

	reset_Vec( &self->opaqueXf );
	reset_Vec( &self->transparentXf );
	splitOpaqueTransparent( &self->opaqueXf, &self->transparentXf, tn_list );

	//reset_Vec( &self->flaresXf );
	//collectFlares( &self->flaresXf, tn_list );

	calculateWorldCoordinates(
		getElement_Vec( &self->cpuMatrixes, 0 ),
		tn_list
	);

	handleLighting( &self->dl, tn_list );

	applyAdditionalTransformations(
		getCamera_Engine( self ),
		tn_list
	);

	Vec const* vec[] = { &self->opaqueXf, &self->transparentXf };
	buildSceneDL(
		&self->dl,
		getCamera_Engine( self ),
		vec, sizeof( vec ) / sizeof( *vec ),
		&self->materials
	);
}

// DANGER: must be invoked AFTER sceneTaskF already rendered a scene
void flareTaskF_Engine( Engine* self, TreeNode* scene ) {

	static Raw_Xformer root;
	static bool initialized = false;
	if( !initialized ) {
		root = createRaw_Engine( self );
		initialized = true;
	}

	////////
	static TreeNode const* tn_list[1+MAX_NODES];
	flattenTree( scene, tn_list );

	reset_Vec( &self->flaresXf );
	collectFlares( &self->flaresXf, tn_list );

	///////
	GfxDevice const* gdev = getGfxDevice_Sys();
	float aspect_ratio = getAspectRatio_Surface(
		getSurface_GfxDevice( gdev, SURFACE_COLOR_DRAW )
	);


	TreeNode root_node, /*camera_node,*/ flare_node[MAX_NODES];
	init_TreeNode( &root_node, &root );
	///init_TreeNode( &camera_node, getCamera_Engine( self ) );

	VecIterator flares_iterator = createIterator_Vec( &self->flaresXf );
	Flare_Xformer** flare_xf;
	usize_t fnode_i = 0;
	while( (flare_xf = dereference_VecIterator( &flares_iterator )) ) {
		VecIterator pairs_iterator = createIterator_Vec( (*flare_xf)->meshes );
		Pair_StaticMesh_Xformer_Scale* pair_xf_s;
		usize_t const pairs_num = getLength_Vec( (*flare_xf)->meshes );
		usize_t pair_i = 0;

		Matrix baseMtx = *getResultMatrix_Xformer( (*flare_xf)->xf );
		while( (pair_xf_s = dereference_VecIterator( &pairs_iterator )) ) {
			init_TreeNode( &flare_node[fnode_i], &(pair_xf_s->static_mesh_xf) );

			// calculate light position in clip coordinates
			Matrix mtx;
			if( 0 == pair_i ) {
				mtx = baseMtx;
			} else {
				multiply_Matrix( &mtx,
					&baseMtx,
					getMatrix_Xformer( pair_xf_s->static_mesh_xf.xf )
				);
			}
			multiply_Matrix( &mtx,
				&mtx,
				getProjectionMatrix_Engine( self )								// DANGER: assuming it is multiplied by the inverse view!!!
			);

			// clip coords -> NDC
			Vector3f light_pos_ndc;
			init_Vector3f( &light_pos_ndc,
				MATRIX_GET(mtx, 3, 0) / MATRIX_GET(mtx, 3, 3),
				MATRIX_GET(mtx, 3, 1) / MATRIX_GET(mtx, 3, 3),
				MATRIX_GET(mtx, 3, 2) / MATRIX_GET(mtx, 3, 3)											// ~198 / 200
			);
			if( (0 == pair_i) && (
				(light_pos_ndc.x <= -1.0f || light_pos_ndc.x >= 1.0f) ||
				(light_pos_ndc.y <= -1.0f || light_pos_ndc.y >= 1.0f) ||
				(light_pos_ndc.z <= -1.0f || light_pos_ndc.z >= 1.0f)
			) ) {
				goto next;
			}

			// position in device coordinates
			Vector3f light_pos_scr;
			init_Vector3f( &light_pos_scr,
				0.5f*getWidth_Viewport( getViewport_Engine( self ) ) * (light_pos_ndc.x + 1.0f),
				-0.5f*getHeight_Viewport( getViewport_Engine( self ) ) * (light_pos_ndc.y - 1.0f),
				0.0f
			);

			// check if occluded
			Surface const* zbuf = getZBuffer_Engine( self );
			float zval = -1.0f;
			if( coordInBounds_Surface( zbuf, light_pos_scr.x, light_pos_scr.y ) ) { 
				uint16_t zdev = *((uint16_t*)getValue_Surface( zbuf, light_pos_scr.x, light_pos_scr.y ));
				zval = zAsFloat_GfxDevice( gdev, zdev );
			}
			if( false || (light_pos_ndc.z < zval ) ) {
				attachChild_TreeNode( &root_node, &flare_node[fnode_i] );

				// calculate final flare
				Vector3f final_pos_ndc;
				init_Vector3f( &final_pos_ndc, -light_pos_ndc.x, -light_pos_ndc.y, light_pos_ndc.z );

				// calculate direction
				Vector3f dir;
				sub_Vector3f( &dir, &final_pos_ndc, &light_pos_ndc );

				// calculate intermediate flare
				Vector3f curr_pos_ndc;
				if( 1 == pairs_num ) {
					curr_pos_ndc = light_pos_ndc;
				} else {
					scale_Vector3f( &curr_pos_ndc, &dir, pair_i * (1.0f / (pairs_num - 1)) );
					add_Vector3f( &curr_pos_ndc, &curr_pos_ndc, &light_pos_ndc );
				}

				// transform
				Matrix mv;
				scale_Matrix( &mv, pair_xf_s->scale, aspect_ratio*pair_xf_s->scale, 1.0f );
				transformLocally_Xformer( pair_xf_s->static_mesh_xf.xf, &mv );

				translate_Matrix( &mv, curr_pos_ndc.x, curr_pos_ndc.y, curr_pos_ndc.z );
				transformLocally_Xformer( pair_xf_s->static_mesh_xf.xf, &mv );
			}
next:
			++pair_i;
			++fnode_i;
			next_VecIterator( &pairs_iterator );
		}

		destroy_VecIterator( &pairs_iterator );
		next_VecIterator( &flares_iterator );
	}
	destroy_VecIterator( &flares_iterator );

	// render
	Camera_Xformer const* camera_xf = getCamera_Engine( self );
	setCamera_Engine( self, &self->camera_2d_xf );
	setParams_OrthoCamera(
		getOrtho_Camera( getCamera_Engine( self )->camera ),
		-1.0f, 1.0f, -1.0f, 1.0f, 1.33f, 1000.0f
	);

	sceneTaskF_Engine( self, &root_node );

	setCamera_Engine( self, camera_xf );

	// cleanup
	flares_iterator = createIterator_Vec( &self->flaresXf );
	fnode_i = 0;
	while( (flare_xf = dereference_VecIterator( &flares_iterator )) ) {
		for( usize_t pair_i = 0; pair_i < getLength_Vec( (*flare_xf)->meshes ); ++pair_i ) {
			destroy_TreeNode( &flare_node[fnode_i++] );
		}
		next_VecIterator( &flares_iterator );
	}
	destroy_VecIterator( &flares_iterator );

	///destroy_TreeNode( &camera_node );
	destroy_TreeNode( &root_node );
}

void clearColorbufF_Engine( Engine* self, void* arg ) {
	Vector3f const* color = arg;
	pushSetClearColor_DList( &self->dl,
		color->x,
		color->y,
		color->z,
		1.0f
	);
	pushClearBuffers_DList( &self->dl, ColorBuffer );
}

void clearZbufF_Engine( Engine* self, void* arg ) {
	float const* zval = arg;
	pushSetClearZValue_DList( &self->dl, *zval );
	pushClearBuffers_DList( &self->dl, ZBuffer );
}

void processTask_Engine( Engine* self, Task const* task ) {
	Surface const* cbuf = getRenderTarget_Engine( self );
	Surface const* zbuf = getZBuffer_Engine( self );

	reset_DList( &self->dl );
	pushStart_DList( &self->dl );
	pushSetViewport_DList( &self->dl, getViewport_Engine( self ) );
	pushSetBuffer_DList(
		&self->dl,
		ColorBuffer,
		0,
		getWidth_Surface( cbuf ),
		getHeight_Surface( cbuf ),
		getData_Surface( cbuf )
	);
	pushSetBuffer_DList(
		&self->dl,
		ZBuffer,
		0,
		getWidth_Surface( zbuf ),
		getHeight_Surface( zbuf ),
		getData_Surface( getZBuffer_Engine( self ) )
	);

	task->f( self, task->arg );

	pushSync_DList( &self->dl );
	pushEnd_DList( &self->dl );
}

Xformer* createXformer_Engine( Engine* self, ProcessFunc_Xformer processor ) {
	Xformer xf;
	init_Xformer( &xf, processor );
	pushBack_Vec( &self->xformers, &xf );

	return getElement_Vec(
		&self->xformers,
		getLength_Vec( &self->xformers ) - 1
	);
}

Raw_Xformer createRaw_Engine( Engine* self ) {
	Raw_Xformer raw_xf;

	raw_xf.tag = OBJECT_RAW;
	raw_xf.xf = createXformer_Engine( self, NULL );

	return raw_xf;
}

StaticMesh_Xformer createStaticMesh_Engine( Engine* self, StaticMesh const* staticMesh ) {
	StaticMesh_Xformer staticMesh_xf;

	pushBack_Vec( &self->static_meshes, staticMesh );
	staticMesh_xf.tag = OBJECT_STATICMESH;
	staticMesh_xf.static_mesh = getElement_Vec(
		&self->static_meshes,
		getLength_Vec( &self->static_meshes ) - 1
	);

	staticMesh_xf.xf = createXformer_Engine( self, NULL );

	return staticMesh_xf;
}

Mesh_Xformer createMesh_Engine( Engine* self ) {
	Mesh_Xformer mesh_xf;
	Mesh dummy;

	pushBack_Vec( &self->meshes, &dummy );
	mesh_xf.tag = OBJECT_MESH;
	mesh_xf.mesh = getTail_Vec(
		&self->meshes
	);

	mesh_xf.xf = createXformer_Engine( self, NULL );

	return mesh_xf;
}

Billboard_Xformer createBillboard_Engine( Engine* self, StaticMesh const* staticMesh ) {
	Billboard_Xformer bb_xf;

	Billboard bb;
	init_Billboard( &bb, staticMesh );
	pushBack_Vec( &self->billboards, &bb );

	bb_xf.tag = OBJECT_BILLBOARD;
	bb_xf.billboard = getElement_Vec(
		&self->billboards,
		getLength_Vec( &self->billboards ) - 1
	);

	bb_xf.xf = createXformer_Engine( self, processBillboard_Xformer );

	return bb_xf;
}

Camera_Xformer createCamera_Engine( Engine* self, CameraType type ) {
	Camera_Xformer cam_xf;

	Camera cam;
	init_Camera( &cam, type );
	pushBack_Vec( &self->cameras, &cam );

	cam_xf.tag = OBJECT_CAMERA;
	cam_xf.camera = getElement_Vec(
		&self->cameras,
		getLength_Vec( &self->cameras ) - 1
	);

	cam_xf.xf = createXformer_Engine( self, processCamera_Xformer );

	return cam_xf;
}

Lux_Xformer createLux_Engine( Engine* self ) {
	Lux_Xformer lux_xf;

	Lux lux;
	init_Lux( &lux );
	pushBack_Vec( &self->luxes, &lux );
	lux_xf.lux = getElement_Vec(
		&self->luxes,
		getLength_Vec( &self->luxes ) - 1
	);

	lux_xf.tag = OBJECT_LUX;
	lux_xf.xf = createXformer_Engine( self, NULL );

	return lux_xf;
}

Flare_Xformer createFlare_Engine( Engine* self, Vec* flare_meshes ) {
	Flare_Xformer flare_xf;

	flare_xf.tag = OBJECT_FLARE;

	flare_xf.xf = createXformer_Engine( self, NULL );
	flare_xf.meshes = flare_meshes;

	return flare_xf;
}

