#include <abort.h>
#include <sys/gfxdevice.h>
#include <ultra64.h>

#define BUFFER_W		320
#define BUFFER_H		240
#define	FRAMEBUFFER0	((void*)0x80200000)
#define	FRAMEBUFFER1	(void*)((uintptr_t)FRAMEBUFFER0 + BUFFER_W*BUFFER_H*2)
#define	AUXBUFFER		(void*)((uintptr_t)FRAMEBUFFER1 + BUFFER_W*BUFFER_H*2)
#define ZBUFFER			((void*)0x80300000)

static GfxMode const gfx_modes[] = {
	{ 320, 240, GFXCOLORMODE_RGBA_5551 }
};

static void waitForVBI_Sys( GfxDevice* self ) {
	if( MQ_IS_FULL( &self->viMessageQ ) ) {
		osRecvMesg(
			&self->viMessageQ,
			NULL,
			OS_MESG_BLOCK
		);
	}
	osRecvMesg(
		&self->viMessageQ,
		NULL,
		OS_MESG_BLOCK
	);
}


void init_GfxDevice( GfxDevice* self ) {
	self->isRendering = false;

	osCreateViManager(
		OS_PRIORITY_VIMGR
	);
	osViSetSpecialFeatures(
		OS_VI_GAMMA_DITHER_OFF
	);

	setMode_GfxDevice( self, 0 );

	/* Create message queue for VI; interestingly, a primitive drawn blinks without it */
	osCreateMesgQueue(
		&self->viMessageQ,
		&self->viMessages,
		1
	);
	//osViSetEvent(&sysState.viMessageQ, NULL, 1);
	osCreateMesgQueue(
		&self->vbiMessageQ,
		&self->vbiMessages,
		1
	);
	osViSetEvent(
		&self->vbiMessageQ,
		NULL,
		1
	);

	/* Message queues for RSP/RDP */
	osCreateMesgQueue(
		&self->rspMessageQ,
		&self->rspMessages,
		1
	);
	osSetEventMesg(
		OS_EVENT_SP,
		&self->rspMessageQ,
		NULL
	);
	osCreateMesgQueue(
		&self->rdpMessageQ,
		&self->rdpMessages,
		1
	);
	osSetEventMesg(
		OS_EVENT_DP,
		&self->rdpMessageQ,
		NULL
	);

	/* Surfaces */
	init_Surface( &self->colorBuffer0,
		PIXELFMT_RGBA_5551,
		FRAMEBUFFER0,
		BUFFER_W,
		BUFFER_H
	);
	init_Surface( &self->colorBuffer1,
		PIXELFMT_RGBA_5551,
		FRAMEBUFFER1,
		BUFFER_W,
		BUFFER_H
	);
	init_Surface( &self->auxBuffer,
		PIXELFMT_RGBA_5551,
		AUXBUFFER,
		BUFFER_W,
		BUFFER_H
	);
	init_Surface( &self->zBuffer,
		PIXELFMT_16,
		ZBUFFER,
		BUFFER_W,
		BUFFER_H
	);
	self->colorBuffer = &self->colorBuffer0;

	init_DLTrans( &self->dltrans );
	init_NativeDL( &self->displayList );
}

Surface const* getSurface_GfxDevice( GfxDevice const* self, SurfaceType type ) {
	switch( type ) {
		case SURFACE_COLOR_DRAW:
			return self->colorBuffer;
		case SURFACE_Z:
			return &self->zBuffer;
		case SURFACE_AUX:
			return &self->auxBuffer;
		default:
			return NULL;
	}
}

void swapBuffers_GfxDevice( GfxDevice* self ) {
	osViSwapBuffer(
		getData_Surface( self->colorBuffer )
	);

	waitForVBI_Sys( self );

	/* change surface pointer */
	if( self->colorBuffer == &self->colorBuffer0 ) {
		self->colorBuffer = &self->colorBuffer1;
	} else {
		self->colorBuffer = &self->colorBuffer0;
	}
}


uint32_t getModeIndex_GfxDevice( GfxDevice const* self ) {
	return self->modeIdx;
}

void setMode_GfxDevice( GfxDevice* self, uint32_t mode_idx ) {
	if( 0 == mode_idx ) {
		osViSetMode(
			&osViModePalLpn1
		);
	} else {
		abort();
	}

	self->modeIdx = mode_idx;
}

GfxMode const* enumerateModes_GfxDevice( GfxDevice const* self ) {
	if( self ) {}

	return gfx_modes;
}

void waitForVBI_GfxDevice( GfxDevice const* self ) {
	GfxDevice* self_mut = (GfxDevice*)self;

	if( MQ_IS_FULL( &self->vbiMessageQ ) ) {
		osRecvMesg(
			&self_mut->vbiMessageQ,
			NULL,
			OS_MESG_BLOCK
		);
	}
	osRecvMesg(
		&self_mut->vbiMessageQ,
		NULL,
		OS_MESG_BLOCK
	);
}

static void waitForRSP( GfxDevice const* self ) {
	GfxDevice* self_mut = (GfxDevice*)self;
	
	osRecvMesg(
		&self_mut->rspMessageQ,
		NULL,
		OS_MESG_BLOCK
	);
}

static void waitForRDP( GfxDevice const* self ) {
	GfxDevice* self_mut = (GfxDevice*)self;

	osRecvMesg(
		&self_mut->rdpMessageQ,
		NULL,
		OS_MESG_BLOCK
	);
}

inline bool isRendering_GfxDevice( GfxDevice const* self ) {
	return self->isRendering;
}

inline void waitForGPU_GfxDevice( GfxDevice const* self ) {
	if( !isRendering_GfxDevice( self ) ) {
		return;
	}

	waitForRDP( self );
	waitForRSP( self );

	GfxDevice* self_mut = (GfxDevice*)self;
	self_mut->isRendering = false;
}

inline NativeDL const* getNativeDL_GfxDevice( GfxDevice const* self ) {
	return &self->displayList;
}

void runGPU_GfxDevice( GfxDevice const* self, DList const* dlist ) {
	GfxDevice* self_mut = (GfxDevice*)self;

	resetCounter_NativeDL( &self_mut->displayList );
	process_DLTrans(
		&self_mut->dltrans,
		dlist,
		&self_mut->displayList
	);

	OSTask_t* p = &self_mut->tlist.t;
	p->type = M_GFXTASK;
	p->flags = OS_TASK_DP_WAIT;
	p->ucode_size = SP_UCODE_SIZE;
	p->ucode_data_size = SP_UCODE_DATA_SIZE;
	p->dram_stack = &self_mut->dram_stack[0];
	p->dram_stack_size = SP_DRAM_STACK_SIZE8;
	p->ucode = (u64*)gspF3DEX2_xbusTextStart;
	p->ucode_data = (u64*)gspF3DEX2_xbusDataStart;
	p->ucode_boot = (u64*)rspbootTextStart;
	p->ucode_boot_size = (u32)rspbootTextEnd - (u32)rspbootTextStart;
	p->data_ptr = (u64*)getHead_NativeDL( &self_mut->displayList );
	p->data_size = getLength_NativeDL( &self->displayList ) * sizeof(Gfx);
	p->output_buff = NULL;
	p->output_buff_size = 0;
	p->yield_data_ptr = self_mut->yieldBuffer;
	p->yield_data_size = OS_YIELD_DATA_SIZE;

	osWritebackDCacheAll();
	self_mut->isRendering = true;
	osSpTaskStart( &self_mut->tlist );
}

float zAsFloat_GfxDevice( GfxDevice const* self, uint32_t zval ) {
	if( self ) {}

	static struct {
		uint32_t shift;
		uint32_t add;
	} z_tab[] = {
		{ 6, 0x00000 },
		{ 5, 0x20000 },
		{ 4, 0x30000 },
		{ 3, 0x38000 },
		{ 2, 0x3C000 },
		{ 1, 0x3E000 },
		{ 0, 0x3F000 },
		{ 0, 0x3F800 },
	};

	zval >>= 2;
	uint16_t exponent = zval >> 11;
	uint16_t mantissa = zval & 0x7FF;
	int32_t decompressed = (mantissa << z_tab[exponent].shift) + z_tab[exponent].add;

	return ((float)((decompressed >> (1+2)) - 0x3FE0)) / 0x3FE0;
}
